import swapper
from baseapp_core.graphql import Node, get_object_type_for_model

ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")


class ChatsQueries:
    chat_room = Node.Field(get_object_type_for_model(ChatRoom))



from baseapp_core.tests.settings import *  # noqa
from django.utils.translation import gettext_lazy as _

# Application definition
INSTALLED_APPS += [
    "channels",
    "avatar",
    "trench",
    "rest_framework_simplejwt",
    "testproject.testapp",
    "push_notifications",
    "baseapp_auth",
    "baseapp_profiles",
    "baseapp_reactions",
    "baseapp_reports",
    "baseapp_pages",
    "baseapp_comments",
    "baseapp_follows",
    "baseapp_blocks",
    "baseapp_chats",
    "baseapp_notifications",
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "baseapp_profiles.middleware.CurrentProfileMiddleware",
]

GRAPHENE["MIDDLEWARE"] = (
    "baseapp_profiles.graphql.middleware.CurrentProfileMiddleware",
) + GRAPHENE["MIDDLEWARE"]

LANGUAGE_CODE = "en"
LANGUAGES = [("en", _("English")), ("es", _("Spanish")), ("pt", _("Portuguese"))]

BROKER_URL = "memory:///"
CELERY_BROKER_URL = "memory:///"
CELERY_RESULT_BACKEND = "cache+memory:///"
BROKER_BACKEND = "memory"
CELERY_TASK_ALWAYS_EAGER = True
CELERY_TASK_EAGER_PROPAGATES = True

ROOT_URLCONF = "testproject.urls"

# Auth
AUTH_USER_MODEL = "testapp.User"

AUTHENTICATION_BACKENDS = [
    "django.contrib.auth.backends.ModelBackend",
    "baseapp_auth.permissions.UsersPermissionsBackend",
    "baseapp_comments.permissions.CommentsPermissionsBackend",
    "baseapp_profiles.permissions.ProfilesPermissionsBackend",
    "baseapp_chats.permissions.ChatsPermissionsBackend",
]

BASEAPP_COMMENTS_ENABLE_NOTIFICATIONS = False
BASEAPP_REACTIONS_ENABLE_NOTIFICATIONS = False


NOTIFICATIONS_NOTIFICATION_MODEL = "baseapp_notifications.Notification"

# Baseapp Auth
BASEAPP_AUTH_USER_FACTORY = "testproject.testapp.tests.factories.UserFactory"



import textwrap

import pytest
import swapper
from baseapp_core.tests.factories import UserFactory
from channels.db import database_sync_to_async

from ..utils import send_message
from .factories import ChatRoomFactory, ChatRoomParticipantFactory

Message = swapper.load_model("baseapp_chats", "Message")
Verbs = Message.Verbs

pytestmark = pytest.mark.django_db(transaction=True)


@pytest.mark.asyncio
async def test_user_recieves_news_message_subscription_event(
    django_user_client, graphql_ws_user_client
):
    room = await database_sync_to_async(ChatRoomFactory)(created_by=django_user_client.user)
    await database_sync_to_async(ChatRoomParticipantFactory)(
        profile=django_user_client.user.profile, room=room
    )
    user = await database_sync_to_async(UserFactory)()
    await database_sync_to_async(ChatRoomParticipantFactory)(room=room, profile=user.profile)

    # Establish & initialize WebSocket GraphQL connection.
    client = await graphql_ws_user_client(consumer_attrs={"strict_ordering": True})

    # Subscribe to GraphQL subscription.
    sub_id = await client.send(
        msg_type="subscribe",
        payload={
            "query": textwrap.dedent(
                """
                subscription op_name($roomId: ID!) {
                  chatRoomOnNewMessage(roomId: $roomId) {
                    message {
                      node {
                        id
                        content
                      }
                    }
                  }
                }
                """
            ),
            "variables": {"roomId": room.relay_id},
            "operationName": "op_name",
        },
    )

    await client.assert_no_messages()

    message = await database_sync_to_async(send_message)(
        profile=user.profile,
        content="Hi!",
        room=room,
        user=user,
        verb=Verbs.SENT_MESSAGE,
    )

    # Check that subscription message were sent.
    resp = await client.receive(assert_id=sub_id, assert_type="next")
    assert resp["data"]["chatRoomOnNewMessage"]["message"]["node"]["content"] == message.content

    # Disconnect and wait the application to finish gracefully.
    await client.finalize()


@pytest.mark.asyncio
async def test_build_absolute_uri_on_graphql_subscription(
    django_user_client, graphql_ws_user_client, image_djangofile
):
    room = await database_sync_to_async(ChatRoomFactory)(created_by=django_user_client.user)
    await database_sync_to_async(ChatRoomParticipantFactory)(
        profile=django_user_client.user.profile, room=room
    )

    user = await database_sync_to_async(UserFactory)()
    user.profile.image = image_djangofile
    await user.profile.asave()
    await database_sync_to_async(ChatRoomParticipantFactory)(profile=user.profile, room=room)

    # Establish & initialize WebSocket GraphQL connection.
    client = await graphql_ws_user_client(consumer_attrs={"strict_ordering": True})

    # Subscribe to GraphQL subscription.
    sub_id = await client.send(
        msg_type="subscribe",
        payload={
            "query": textwrap.dedent(
                """
                subscription op_name($roomId: ID!) {
                  chatRoomOnNewMessage(roomId: $roomId) {
                    message {
                      node {
                        profile {
                          image(width: 100, height: 100) {
                            url
                          }
                        }
                      }
                    }
                  }
                }
                """
            ),
            "variables": {"roomId": room.relay_id},
            "operationName": "op_name",
        },
    )

    await client.assert_no_messages()

    # trigger a subscription message
    await database_sync_to_async(send_message)(
        user=user,
        profile=user.profile,
        content="Hi!",
        room=room,
        verb=Verbs.SENT_MESSAGE,
    )

    # Check that subscription message were sent.
    resp = await client.receive(assert_id=sub_id, assert_type="next")
    assert resp["data"]["chatRoomOnNewMessage"]["message"]["node"]["profile"]["image"][
        "url"
    ].startswith("http")

    # Disconnect and wait the application to finish gracefully.
    await client.finalize()


@pytest.mark.asyncio
async def test_user_recieves_message_count_update(django_user_client, graphql_ws_user_client):
    room = await database_sync_to_async(ChatRoomFactory)(created_by=django_user_client.user)
    await database_sync_to_async(ChatRoomParticipantFactory)(
        profile=django_user_client.user.profile, room=room
    )
    user = await database_sync_to_async(UserFactory)()
    await database_sync_to_async(ChatRoomParticipantFactory)(room=room, profile=user.profile)

    # Establish & initialize WebSocket GraphQL connection.
    client = await graphql_ws_user_client(consumer_attrs={"strict_ordering": True})

    # Subscribe to GraphQL subscription.
    sub_id = await client.send(
        msg_type="subscribe",
        payload={
            "query": textwrap.dedent(
                """
                subscription op_name($profileId: ID!) {
                  chatRoomOnMessagesCountUpdate(profileId: $profileId) {
                    profile {
                      ... on ChatRoomsInterface {
                        unreadMessagesCount
                      }
                    }
                  }
                }
                """
            ),
            "variables": {"profileId": django_user_client.user.profile.relay_id},
            "operationName": "op_name",
        },
    )

    await client.assert_no_messages()

    await database_sync_to_async(send_message)(
        user=user,
        profile=user.profile,
        content="Hi!",
        room=room,
        verb=Verbs.SENT_MESSAGE,
    )

    # Check that subscription message were sent.

    resp = await client.receive(assert_id=sub_id, assert_type="next")

    # Disconnect and wait the application to finish gracefully.
    await client.finalize()

    assert resp["data"]["chatRoomOnMessagesCountUpdate"]["profile"]["unreadMessagesCount"] == 1



# Generated by Django 5.0.8 on 2025-01-10 12:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("baseapp_chats", "0007_messagestatus_increment_unread_count"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="chatroomparticipant",
            options={"ordering": ["-role", "profile__name"]},
        ),
    ]



# Generated by Django 5.0.9 on 2024-12-09 14:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("baseapp_chats", "0003_alter_unreadmessagecount_room"),
    ]

    operations = [
        migrations.AddField(
            model_name="chatroomparticipant",
            name="has_archived_room",
            field=models.BooleanField(default=False),
        ),
    ]



import factory


class ChatRoomFactory(factory.django.DjangoModelFactory):
    created_by = factory.SubFactory("baseapp_core.tests.factories.UserFactory")

    class Meta:
        model = "baseapp_chats.ChatRoom"


class MessageFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory("baseapp_core.tests.factories.UserFactory")
    room = factory.SubFactory(ChatRoomFactory)

    class Meta:
        model = "baseapp_chats.Message"


class ChatRoomParticipantFactory(factory.django.DjangoModelFactory):
    room = factory.SubFactory(ChatRoomFactory)
    profile = factory.SubFactory("baseapp_profiles.tests.factories.ProfileFactory")

    class Meta:
        model = "baseapp_chats.ChatRoomParticipant"


class MessageStatusFactory(factory.django.DjangoModelFactory):
    message = factory.SubFactory(MessageFactory)
    profile = factory.SubFactory("baseapp_profiles.tests.factories.ProfileFactory")

    class Meta:
        model = "baseapp_chats.MessageStatus"



import swapper
from django.contrib.auth.backends import BaseBackend
from django.db import models

ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
Message = swapper.load_model("baseapp_chats", "Message")
ChatRoomParticipant = swapper.load_model("baseapp_chats", "ChatRoomParticipant")
Block = swapper.load_model("baseapp_blocks", "Block")
Profile = swapper.load_model("baseapp_profiles", "Profile")


class ChatsPermissionsBackend(BaseBackend):
    def can_add_chatroom(self, user_obj, obj):
        if isinstance(obj, dict):
            participants = obj["participants"]
            current_profile = obj["profile"]

            if len(participants) < 2:
                return False

            if current_profile:
                my_profile_ids = [current_profile.id]
            else:
                my_profile_ids = Profile.objects.filter_user_profiles(user_obj).values_list(
                    "id", flat=True
                )

            participant_profile_ids = [participant.pk for participant in participants]

            blocks_qs = Block.objects.filter(
                models.Q(actor_id__in=my_profile_ids, target_id__in=participant_profile_ids)
                | models.Q(actor_id__in=participant_profile_ids, target_id__in=my_profile_ids)
            )

            if blocks_qs.exists():
                return False

            return True

    def can_modify_chatroom(self, user_obj, obj):
        if isinstance(obj, dict):
            room = obj["room"]
            add_participants = obj["add_participants"]
            current_profile = obj["profile"]

            if not getattr(room, "is_group", False):
                return False

            chat_room_participant = ChatRoomParticipant.objects.filter(
                room=room, profile_id=current_profile.pk
            ).first()
            if (
                not chat_room_participant
                or not chat_room_participant.role
                == ChatRoomParticipant.ChatRoomParticipantRoles.ADMIN
            ):
                return False

            participant_profile_ids = [participant.pk for participant in add_participants]

            blocks_qs = Block.objects.filter(
                models.Q(actor_id=current_profile.id, target_id__in=participant_profile_ids)
                | models.Q(actor_id__in=participant_profile_ids, target_id=current_profile.id)
            )

            if blocks_qs.exists():
                return False

            return True

    def has_perm(self, user_obj, perm, obj=None):
        if perm == "baseapp_chats.add_chatroom" and user_obj.is_authenticated:
            return self.can_add_chatroom(user_obj, obj)
        if perm == "baseapp_chats.add_chatroom_with_profile":
            return user_obj.has_perm("baseapp_profiles.use_profile", obj)
        if perm == "baseapp_chats.delete_chat":
            if isinstance(obj, ChatRoom):
                return obj.user_id == user_obj.id or user_obj.has_perm(
                    "baseapp_profiles.use_profile", obj.actor
                )
        if perm == "baseapp_chats.modify_chatroom":
            return self.can_modify_chatroom(user_obj, obj)
        if perm == "baseapp_chats.view_chatroom":
            my_profile_ids = Profile.objects.filter_user_profiles(user_obj).values_list(
                "id", flat=True
            )

            if not obj.participants.filter(models.Q(profile_id__in=my_profile_ids)).exists():
                return False

            participant_profile_ids = obj.participants.values_list("profile_id", flat=True)

            blocks_qs = Block.objects.filter(
                models.Q(actor_id__in=my_profile_ids, target_id__in=participant_profile_ids)
                | models.Q(actor_id__in=participant_profile_ids, target_id__in=my_profile_ids)
            )

            if blocks_qs.exists():
                return False

            return obj.participants.filter(profile_id__in=my_profile_ids).exists()

        if perm == "baseapp_chats.list_chatrooms":
            return obj.check_if_member(user_obj)

        if perm == "baseapp_chats.add_message":
            current_profile = obj["profile"]
            room = obj["room"]

            if current_profile:
                my_profile_ids = [current_profile.id]
            else:
                my_profile_ids = Profile.objects.filter_user_profiles(user_obj).values_list(
                    "id", flat=True
                )

            participant_profile_ids = room.participants.values_list("profile_id", flat=True)

            blocks_qs = Block.objects.filter(
                models.Q(actor_id__in=my_profile_ids, target_id__in=participant_profile_ids)
                | models.Q(actor_id__in=participant_profile_ids, target_id__in=my_profile_ids)
            )

            if blocks_qs.exists():
                return False

            return room.participants.filter(profile_id__in=my_profile_ids).exists()

        if perm == "baseapp_chats.change_message" and user_obj.is_authenticated:
            profile = obj.get("profile", None)
            message = obj.get("message", None)
            if profile and message and isinstance(message, Message):
                return profile.id == message.profile.id

            return False



import graphene
import swapper
from baseapp_core.graphql import get_object_type_for_model
from django.db.models import Sum
from graphene import relay
from graphene_django.filter import DjangoFilterConnectionField

ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
UnreadMessageCount = swapper.load_model("baseapp_chats", "UnreadMessageCount")
Block = swapper.load_model("baseapp_blocks", "Block")


class ChatRoomsInterface(relay.Node):
    chat_rooms = DjangoFilterConnectionField(get_object_type_for_model(ChatRoom))
    unread_messages_count = graphene.Int()

    def resolve_chat_rooms(self, info, **kwargs):
        if not info.context.user.has_perm("baseapp_chats.list_chatrooms", self):
            return ChatRoom.objects.none()

        qs = ChatRoom.objects.filter(
            participants__profile_id=self.pk,
        ).order_by("-last_message_time", "-created")

        # Exclude rooms with any blocked participant
        blocking_profile_ids = Block.objects.filter(actor_id=self.pk).values_list(
            "target_id", flat=True
        )
        qs = qs.exclude(participants__profile_id__in=blocking_profile_ids)

        # Exclude rooms with any participant that blocks the profile (self)
        blocker_profile_ids = Block.objects.filter(target_id=self.pk).values_list(
            "actor_id", flat=True
        )
        qs = qs.exclude(participants__profile_id__in=blocker_profile_ids)

        return qs

    def resolve_unread_messages_count(self, info, **kwargs):
        if not info.context.user.has_perm("baseapp_chats.list_chatrooms", self):
            return None

        aggregate_result = UnreadMessageCount.objects.filter(
            profile_id=self.pk,
        ).aggregate(total_count=Sum("count"))

        return aggregate_result["total_count"] or 0



import django_filters
import swapper
from baseapp_core.graphql import get_pk_from_relay_id
from django.db.models import Q

ChatRoom = swapper.load_model("baseapp_chats", "ChatRoom")
ChatRoomParticipant = swapper.load_model("baseapp_chats", "ChatRoomParticipant")


class ChatRoomFilter(django_filters.FilterSet):
    q = django_filters.CharFilter(method="filter_q")
    profile_id = django_filters.CharFilter(method="filter_profile_id")
    unread_messages = django_filters.BooleanFilter(method="filter_unread_messages")
    archived = django_filters.BooleanFilter(method="filter_archived")

    order_by = django_filters.OrderingFilter(fields=(("created", "created"),))

    class Meta:
        model = ChatRoom
        fields = ["q", "order_by"]

    def filter_q(self, queryset, name, value):
        if not value:
            return queryset

        try:
            user_profile_id = self.request.user.current_profile.pk
        except AttributeError:
            user_profile_id = None

        return queryset.filter(
            (Q(is_group=True) & Q(title__icontains=value))
            | (
                Q(is_group=False)
                & Q(participants__profile__name__icontains=value)
                & (
                    Q(participants__profile_id__lt=user_profile_id)
                    | Q(participants__profile_id__gt=user_profile_id)
                )
            )
        ).distinct()

    def filter_profile_id(self, queryset, name, value):
        pk = get_pk_from_relay_id(value)
        return queryset.filter(participants__profile_id=pk)

    def filter_unread_messages(self, queryset, name, value):
        if value:
            profile_id = self.data.get("profile_id", None)
            try:
                user_profile = self.request.user.current_profile
            except AttributeError:
                return queryset.none()

            unread_messages_profile_pk = (
                get_pk_from_relay_id(profile_id) if profile_id else user_profile.pk
            )

            return (
                queryset.prefetch_related("unread_messages")
                .filter(
                    Q(unread_messages__profile_id=unread_messages_profile_pk),
                    Q(unread_messages__count__gt=0) | Q(unread_messages__marked_unread=True),
                )
                .distinct()
            )

        return queryset

    def filter_archived(self, queryset, name, value):
        try:
            user_profile = self.request.user.current_profile
        except AttributeError:
            return queryset.none()

        return queryset.filter(
            participants__profile_id=user_profile.pk, participants__has_archived_room=value
        ).distinct()



from django.apps import AppConfig


class ChatsConfig(AppConfig):
    default = True
    name = "baseapp_chats"
    verbose_name = "BaseApp Chats"
    default_auto_field = "django.db.models.BigAutoField"



import swapper
from baseapp_notifications import send_notification
from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

CONTENT_LINKED_PROFILE_ACTOR = "{content_linked_profile_actor}"
CONTENT_LINKED_PROFILE_TARGET = "{content_linked_profile_target}"

Message = swapper.load_model("baseapp_chats", "Message")
MessageStatus = swapper.load_model("baseapp_chats", "MessageStatus")
UnreadMessageCount = swapper.load_model("baseapp_chats", "UnreadMessageCount")
Verbs = Message.Verbs
MessageType = Message.MessageType

User = get_user_model()


def send_message(
    room,
    profile,
    user,
    content,
    content_linked_profile_actor=None,
    content_linked_profile_target=None,
    room_id=None,
    message_type=MessageType.USER_MESSAGE,
    verb=Verbs.SENT_MESSAGE,
    action_object=None,
    extra_data=None,
    in_reply_to=None,
):
    message = Message.objects.create(
        user=user,
        profile=profile,
        content=content,
        content_linked_profile_actor=content_linked_profile_actor,
        content_linked_profile_target=content_linked_profile_target,
        room=room,
        message_type=message_type,
        verb=verb,
        in_reply_to=in_reply_to,
        action_object=action_object,
        extra_data=extra_data,
    )

    from baseapp_chats.graphql.subscriptions import ChatRoomOnNewMessage

    room.last_message_time = message.created
    room.last_message = message
    room.save()

    ChatRoomOnNewMessage.new_message(room_id=room_id or room.relay_id, message=message)

    return message


def send_new_chat_message_notification(room, message, info):
    for participant in room.participants.all():
        recipients = participant.profile.get_all_users()
        for recipient in recipients:
            send_notification(
                add_to_history=False,
                send_email=False,
                send_push=True,
                sender=message.profile,
                recipient=recipient,
                verb="CHATS.NEW_CHAT_MESSAGE",
                action_object=message,
                target=room,
                level="info",
                description=_("You got a new message"),
                notification_url=f"{settings.FRONT_URL}/chat?roomId={room.relay_id}",
                push_title=_("New message"),
                push_description=_("You got a new message"),
            )



# Generated by Django 5.0.9 on 2024-12-19 08:39

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("baseapp_chats", "0004_chatroomparticipant_has_archived_room"),
    ]

    operations = [
        pgtrigger.migrations.RemoveTrigger(
            model_name="messagestatus",
            name="increment_unread_count",
        ),
        migrations.AddField(
            model_name="unreadmessagecount",
            name="marked_unread",
            field=models.BooleanField(default=False),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="messagestatus",
            trigger=pgtrigger.compiler.Trigger(
                name="increment_unread_count",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition='WHEN (NOT NEW."is_read")',
                    func="\n            INSERT INTO baseapp_chats_unreadmessagecount (room_id, profile_id, marked_unread, count)\n            VALUES ((SELECT room_id FROM baseapp_chats_message WHERE id = NEW.message_id), NEW.profile_id, False, 1)\n            ON CONFLICT (room_id, profile_id)\n            DO UPDATE SET count = baseapp_chats_unreadmessagecount.count + 1;\n            RETURN NULL;\n        ",
                    hash="e18409b9d00a1d881dabf9b7aaa8da101ac5ac28",
                    operation="INSERT",
                    pgid="pgtrigger_increment_unread_count_ec3cd",
                    table="baseapp_chats_messagestatus",
                    when="AFTER",
                ),
            ),
        ),
    ]
